/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * Code for Algorithm 4.1 (TDR Transformed-Density-Rejection), p. 60 
 *
 * this file: algo04_01.c
 *
 * For a demo how to use this algorithm see the file "demo_algo04_01.c".
 *
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "uniform.h"


/* for c=0 */

#define T(aaa) log(aaa)
#define DT(aaa) (1./(aaa))
#define FT(aaa) exp(aaa)
#define FTI(aaa) log(aaa)
#define TI(aaa) exp(aaa)

   static double *p=NULL, *tfp=NULL, *dtfp=NULL, *sp=NULL, *b=NULL, *a=NULL, atotal;
   static double (*pdfh)(double);
   static int NN;

int setup04_01(double (*pdf)(double), double (*dpdf)(double), 
                 double bl, double br, 
                 int N, double constructionp[])
/*
 * Algorithm 4.1 (TDR Transformed-Density-Rejection), p. 60 
 *
 * pdf... function pointer to the density
 * dpdf... function pointer to the derivative of the density
 * bl ... left border of domain
 * br ... right border of domain
 * N  ... number of construction points
 * constructionp[] ... array holding the construction points
 *
 */
{ 
   int i;  

     if(p!=NULL){
       free(p);
       free(tfp);
       free(dtfp);
       free(sp);
       free(b);
       free(a);
     }
     NN = N;
     pdfh =pdf;
     p = malloc((N+1)*sizeof(double));  
     tfp = malloc((N+1)*sizeof(double));  
     dtfp = malloc((N+1)*sizeof(double));  
     sp = malloc((N+1)*sizeof(double));  
     b = malloc((N+1)*sizeof(double));  
     a = malloc((N+1)*sizeof(double));  
     for(i=1;i<=N;i++) p[i] = constructionp[i-1];
     for(i=1;i<=N;i++){
/*1*/  tfp[i] = T((*pdf)(p[i]));
         /* computes the transformed density f-tilda( p_i ) */   
       dtfp[i] = DT((*pdf)(p[i]))*(*dpdf)(p[i]);
         /* computes the derivative of the transformed density f-tilda'( p_i ) */   
       if(i<N){
/*3*/	 b[i] = 0.5*(p[i+1]+p[i]);/*easier than to compute the intersections*/
       }
     }
     b[0] = bl;
     b[N] = br;
     atotal=0.;
     for(i=1;i<=N;i++){
/*2*/  if(i<N)  sp[i]=(tfp[i+1]-tfp[i])/(p[i+1]-p[i]);
/*4*/  a[i] = (FT(tfp[i]+dtfp[i]*(b[i]-p[i]))-
	     FT(tfp[i]+dtfp[i]*(b[i-1]-p[i])))/dtfp[i];
/*5*/  atotal += a[i];
       printf("i%d %f---<%f>---%f area %f\n",i,b[i-1],p[i],b[i],a[i]);
       printf("tfp %f dtfp %f s %f\n",tfp[i],dtfp[i],sp[i]);
     }
     printf("total area below the hat: %f\n",atotal);
     return 1;
}


double generator04_01()
/* is called after the setup to generate random variates */
{   double u, v, x, sx, asum, hx;
   int j;  


     if(p==NULL){
       printf("algo04_01: ERROR you have to call the setup first!!!\n");
       return -9999.;
     }

/*6*/while(1){
/*7*/  v = atotal*uniform();
/*8*/  j = 1;
       asum = a[1];
       while(v > asum){
	 j++;
	 asum += a[j];
       }
/*9*/  v = v - (asum-a[j]);
/*10*/ x = p[j]+(FTI(dtfp[j]*v+FT(tfp[j]+dtfp[j]*(b[j-1]-p[j])))-tfp[j])/dtfp[j];
/*11*/ hx = TI(tfp[j]+dtfp[j]*(x-p[j]));
/*12*/ if(j>1 && x < p[j]) 
/*13*/   sx=TI(tfp[j-1]+sp[j-1]*(x-p[j-1]));
/*14*/ else if(j<NN && x > p[j]) 
/*15*/   sx=TI(tfp[j]+sp[j]*(x-p[j]));
/*16*/ else 
/*17*/   sx=0.;
/*18*/ u = uniform();
/*19*/ if(u*hx <= sx) 
/*20*/   return x;
/*21*/ if(u*hx <= (*pdfh)(x)) 
/*22*/   return x;
     }
}


int free04_01()
/* frees the allocated memory */
{
     free(p);
     p = NULL;
     free(tfp);
     tfp = NULL;
     free(dtfp);
     dtfp = NULL;
     free(sp);
     sp = NULL;
     free(b);
     b = NULL;
     free(a);
     a = NULL;
     return 1;
}


