/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * Code for Algorithm 4.4 (TDR-3-Points (for c = -1/2)), p. 94  
 *
 * this file: algo04_04.c
 *
 * For a demo how to use this algorithm see the file "demo_algo04_04.c".
 *
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "uniform.h"


#define MAX(a,b) ((a)>(b) ? (a) : (b))

    static double fm,tfm,mm,p[4],b[4],ty[4],ts[4],y[4],a[4],A[4],d[4],AT,g;
    static double (*pdfh)(double);


int setup04_04(  double (*pdf)(double),  
                 double bl, double m, double br)
/*
 * Algorithm 4.4 (TDR-3-Points (for c = -1/2)), p. 94  
 *
 * setup ... 1 = new setup; 0 = sample; -1 = free tables
 * pdf... function pointer to the density
 * bl ... left border of domain; bl < -1.e99 is treated as -infinity
 * m  ... mode of the density
 * br ... right border of domain; br > 1.e99 is treated as +infinity
 *
 */
{   double delta;
    int i,j,k;  
    int repeatsetup;

        b[0] = bl;
	b[3] = br;
	mm=m;
	pdfh = pdf;
/*1*/   fm = (*pdf)(m);
        tfm = -1./sqrt(fm);
        g = 0.664;
        repeatsetup = 1;
        while( repeatsetup==1 ){
	  j=1;k=-1;
	  for(i=1; i<=3; i+=2){
	    p[i] = m +k* g/fm;
	    if(k*p[i] < k*b[j+k]){
	      ty[i] = -1./sqrt((*pdf)(p[i])); 
	      ts[i] = (tfm-ty[i])/(m-p[i]);
	      delta = 1.e-5*MAX(fabs(p[i]), fabs(ty[i]/ts[i]));
	      y[i] = -1./sqrt((*pdf)(p[i]-k*delta)); 
	      a[i] = -k*(y[i]-ty[i])/delta;
	      b[j] = p[i]+(tfm-y[i])/a[i];
	      d[i] = y[i]-a[i]*p[i];
	      A[i] = k/(a[i]*tfm);
	      if(fabs(b[j+k])<1.e99) A[i] = A[i]-k/(a[i]*(a[i]*b[j+k]+d[i]));
	    }
	    else{
	      A[i] = 0.;
	      b[j] = b[j+k];
	      if(k*b[j+k] > k* m){
		p[i] = m+(b[j+k]-m)*0.6; 
		ts[i] = (tfm+1./sqrt((*pdf)(p[i])))/(m-p[i]);
	      }
	    }

	    j++;k+=2;
	  }
	  A[2] = (b[2]-b[1])*fm;
	  AT = A[1]+A[2]+A[3];
	  printf("AT %f g%f \n",AT,g);
          repeatsetup = 0;
	  if(AT >= 4.){
	    if(g < 0.666){
	      g = 2.;
	      repeatsetup = 1;
	    }
	    else{
	      printf("AREA below hat for algo04_04 large; perhaps area not close to 1.??\n");
	    }
	  }
	  printf("A1 %f A2 %f A3 %f\n",A[1],A[2],A[3]);
	  printf("p1 %f p3 %f\n", p[1],p[3]);
	  printf("b0 %g b1 %f b2 %f b3 %g\n",b[0],b[1],b[2],b[3]);
	  printf("y1 %f y3 %f\n", y[1],y[3]);
	  printf("ty1 %f ty3 %f\n", ty[1],ty[3]);
	  printf("ts1 %f ts3 %f\n", ts[1],ts[3]);
	  printf("ts1 %f ts3 %f\n", ts[1],ts[3]);
	  printf("a1 %f a3 %f\n", a[1],a[3]);
	  printf("d1 %f d3 %f\n", d[1],d[3]);
	  printf("fm %f tfm %f\n", fm,tfm);
	}
	return 1;
}


double generator04_04()
/* is called after the setup to generate random variates */
{       double U,V,X,lX;
        while(1){
	  U = AT*uniform();
	  if(U <= A[1]){
	    X = -d[1]/a[1]+1./(a[1]*a[1]*(U+1./(a[1]*tfm))); 
	    lX = (a[1]*(U+1./(a[1]*tfm)))*(a[1]*(U+1./(a[1]*tfm)));
	  }
	  else{
	    if(U <= A[1]+A[2]){ 
	      X = (U-A[1])*(b[2]-b[1])/A[2]+b[1];
	      lX = fm;
	    }
	    else{
	      lX = (U-A[1]-A[2]-1./(a[3]*tfm));
	      X = -d[3]/a[3] - 1./(a[3]*a[3]*lX);
	      lX = (a[3]*lX)*(a[3]*lX);
	    }
	  }    
	  V = lX*uniform();
	  if(X < mm){
	    if(X >= p[1] && V*(tfm-(mm-X)*ts[1])*(tfm-(mm-X)*ts[1]) <= 1.){
	      return X;
	    }
	  }
	  else{
	    if(X <= p[3] && V*(tfm-(mm-X)*ts[3])*(tfm-(mm-X)*ts[3])<= 1.){
	      return X;
	    }
	  }
	  if(V <= (*pdfh)(X)){
	    return X;
	  }
	}
}












